// 	WOSA/XRT Excel 5.0 AddIn Version 1.00
//
//	COPYRIGHT:
//
//  	(C) Copyright Microsoft Corp. 1994.  All rights reserved.
//
//   	You have a royalty-free right to use, modify, reproduce and
//   	distribute the Sample Files (and/or any modified version) in
//   	any way you find useful, provided that you agree that
//   	Microsoft has no warranty obligations or liability for any
//   	Sample Application Files which are modified.
//
//
// 	Revision History
//    	October 15, 1994   rc    Version 1.00 Release
//
// 	Comments:
//		Various helper functions which are used throughout project.


#include "precomp.h"
#include "framewrk.h"
#include "guid.h"
#include <initguid.h>

#include "dataobj.h"
#include "requests.h"
#include "request.h"
#include "realtime.h"

                                
/*
 * CreateDispatchInterface
 *
 * Purpose:
 *   Calls CreateStdDispatch to create the standard IDispatch implemenatation
 *   for an object.
 *
 * Parameters:
 *  punkController  	Points to object's IUnknown implementation.
 *  pvProgInterface 	Points to the object's implemenatation of exposed properties and methods. 
 *  ppunkStdDispatch    Returns standard IDispatch implemenatation.
 *  clsid           	Interface id of object in type library. 
 *
 * Return Value:
 *  HRESULT
 *
 */
 
 
HRESULT CreateDispatchInterface(LPUNKNOWN   punkController,
                                LPVOID      pvProgInterface,
                                IUnknown FAR* FAR* ppunkStdDispatch,
                                REFCLSID clsid)
{                          
    LPTYPELIB ptlib = NULL;
    LPTYPEINFO ptinfo = NULL;
    LPUNKNOWN punkStdDispatch = NULL;

	TRY{    
		if(FAILED(CheckPointer((LPVOID) ppunkStdDispatch, 0))){
    		AddInDebugOutput(NULL, TEXT("CreateDispatchInterface")
    			, IDS_DEBUG_BADPOINTER, NULL); 
			return ResultFromScode(E_FAIL);
		}
	
    	*ppunkStdDispatch = NULL;  
    
    	// Note: OLE Automation functions do not set the out pointer parameters 
    	// to NULL on an ResultFromScode(E_FAIL) return as required by the OLE specifications (See
    	// OLE 2.01 specs Section 4.5). This is a problem that will be fixed in a
    	// future version.  In the meantime, do not depend on the out pointer 
    	// parameters being set to NULL when the function fails.
    
    	// Load Type Library. If required, notify user on failure.
    	// parms = libid, major ver, minor ver, locale id, **ITypeLib
    	if (FAILED(LoadRegTypeLib(LIBID_App, 1, 0, LOCALE_SYSTEM_DEFAULT
    			, &ptlib))){
			AddInDebugOutput(NULL
				, TEXT("CreateDispatchInterface")
				, IDS_DEBUG_COULDNOTLOADTYPEINFO
				, NULL); 						   
			MYRAISEEXCEPTION;
    	}
        
	    // Get type information for interface of the object.
    	// parms = object clsid, **ITypeInfo    
    	if (FAILED(ptlib->GetTypeInfoOfGuid(clsid, &ptinfo))){  
			AddInDebugOutput(NULL
				, TEXT("CreateDispatchInterface")
				, IDS_DEBUG_GETTYPEINFOOFGUIDFAILED
				, NULL); 
			MYRAISEEXCEPTION;
	    }   
        
	    // Create a standard implementation of the IDispatch for the object.
    	// parms = outer IUnknown, class to expose, *ITypeInfo, **IUnknown
	    if (FAILED(CreateStdDispatch(punkController, pvProgInterface
	    		, ptinfo, &punkStdDispatch))){
			AddInDebugOutput(NULL
				, TEXT("CreateDispatchInterface")
				, IDS_DEBUG_CREATESTDDISPATCHFAILED
				, NULL); 
			MYRAISEEXCEPTION;
	    }
    
    	ptinfo->Release();
    	ptinfo = NULL;
    	
    	ptlib->Release();
        ptlib = NULL;
        
    	*ppunkStdDispatch = punkStdDispatch; 

    	return S_OK;
	}

	EXCEPT{
		if(ptinfo)
    		ptinfo->Release();

		if(ptlib)
    		ptlib->Release();
	
		if(punkStdDispatch)
			punkStdDispatch->Release();

   		AddInDebugOutput(NULL, TEXT("CreateDispatchInterface")
   			, IDS_DEBUG_EXCEPTION, NULL); 
	
		return ResultFromScode(E_FAIL);
	}
} 

/*
 * LoadTypeInfo
 *
 *  Purpose:
 *   Gets type information of an object's interface from type library.
 *
 * Parameters:
 *  ppunkStdDispatch    Returns type information.
 *  clsid               Interface id of object in type library. 
 *
 * Return Value:
 *  HRESULT
 *
 */
HRESULT LoadTypeInfo(ITypeInfo FAR* FAR* pptinfo, REFCLSID clsid)
{                          
    LPTYPELIB ptlib = NULL;
    LPTYPEINFO ptinfo = NULL;

	TRY{
		if(FAILED(CheckPointer((LPVOID) pptinfo, 0))){
    		AddInDebugOutput(NULL, TEXT("LoadTypeInfo")
    			, IDS_DEBUG_BADPOINTER, NULL); 
			return ResultFromScode(E_FAIL);
		}
	
    	*pptinfo = NULL; 
    
    	// Note: OLE Automation functions do not set the out pointer parameters 
    	// to NULL on an ResultFromScode(E_FAIL) return as required by the OLE specifications (See
    	// OLE 2.01 specs Section 4.5). This is a problem that will be fixed in a
    	// future version.  In the meantime, do not depend on the out pointer 
    	// parameters being set to NULL when the function fails.
    
    	// Load Type Library. If required, notify user on failure.
    	if (FAILED(LoadRegTypeLib(LIBID_App, 1, 0, 0x0409, &ptlib))){
			AddInDebugOutput(NULL
				, TEXT("CreateDispatchInterface")
				, IDS_DEBUG_COULDNOTLOADTYPEINFO
				, NULL); 
			MYRAISEEXCEPTION;
	    }
    
	    // Get type information for interface of the object.      
	    if (FAILED(ptlib->GetTypeInfoOfGuid(clsid, &ptinfo))){
			AddInDebugOutput(NULL
				, TEXT("CreateDispatchInterface")
				, IDS_DEBUG_GETTYPEINFOOFGUIDFAILED
				, NULL); 
			MYRAISEEXCEPTION;
    	}

	    ptlib->Release();
	    ptlib = NULL;
	    
    	*pptinfo = ptinfo;

    	return S_OK;
	}

	EXCEPT{
		if(ptlib)
	    	ptlib->Release();

		if(ptinfo)
			ptinfo->Release();

   		AddInDebugOutput(NULL, TEXT("LoadTypeInfo")
   			, IDS_DEBUG_EXCEPTION, NULL); 

		return ResultFromScode(E_FAIL);
	}
} 
       

/*
 * InvokeMethodHelper
 *
 * Purpose:
 *  Wrapper for IDispatch->GetIDsOfNames and Invoke to hep in calling
 *	methods.
 *
 * Parameters:
 *  pDispatch			IDispatch to use
 *  pdispid			DISPID if already known
 *  pDispParams		params to send
 *  pszMember			Name of Method to call
 *	pvResult			Result of call
 *
 * Return Value:
 *  HRESULT
 *
 */

HRESULT InvokeMethodHelper(LPDISPATCH pDispatch, DISPID FAR *pdispid, 
		DISPPARAMS FAR *pDispParams, LPTSTR pszMember, VARIANT FAR* pvResult){

	TRY{		
		if(FAILED(CheckDispatch(pDispatch))){
    		AddInDebugOutput(NULL, TEXT("CDataObject::Create")
    			, IDS_DEBUG_BADPOINTERTOXRTOBJECT, NULL); 
			return ResultFromScode(E_FAIL);
		}

		if(*pdispid == -1){
			HRESULT hr = pDispatch->GetIDsOfNames(IID_NULL,  (LPTSTR FAR *) &pszMember, 1
				, LOCALE_SYSTEM_DEFAULT, pdispid);
			
			if(FAILED(hr)){
				*pdispid = -1;
				AddInDebugOutput(NULL
					, TEXT("InvokeMethodHelper")
					, IDS_DEBUG_IDISPATCHGETIDSOFNAMESFAILED
					, NULL); 
				MYRAISEEXCEPTION;
			}
		}
			
		return pDispatch->Invoke(*pdispid, IID_NULL
			, LOCALE_SYSTEM_DEFAULT, DISPATCH_METHOD
			, pDispParams, pvResult, NULL ,NULL);
	}

	EXCEPT{
   		AddInDebugOutput(NULL, TEXT("InvokeMethodHelper")
   			, IDS_DEBUG_EXCEPTION, NULL); 
        
		return ResultFromScode(E_FAIL);
	}
}

/*
 * InvokeMethodHelper
 *
 * Purpose:
 *  Wrapper for IDispatch->GetIDsOfNames and Invoke to help in calling
 *	Proeprty Gets
 *
 * Parameters:
 *  pDispatch			IDispatch to use
 *  pdispid			DISPID if already known
 *  pDispParams		params to send
 *  pszMember			Name of Method to call
 *	pvResult			Result of call
 *
 * Return Value:
 *  HRESULT
 *
 */

HRESULT InvokePropGetHelper(LPDISPATCH pDispatch, DISPID FAR *pdispid, 
		DISPPARAMS FAR *pDispParams, LPTSTR pszMember, VARIANT FAR* pvResult){

	TRY{
		if(FAILED(CheckDispatch(pDispatch))){
    		AddInDebugOutput(NULL, TEXT("CDataObject::Create")
    			, IDS_DEBUG_BADPOINTERTOXRTOBJECT, NULL); 
			return ResultFromScode(E_FAIL);
		}

		if(*pdispid == -1){
			HRESULT hr = pDispatch->GetIDsOfNames(IID_NULL,  (LPTSTR FAR *) &pszMember, 1
				, LOCALE_SYSTEM_DEFAULT, pdispid);
			
			if(FAILED(hr)){
				*pdispid = -1;
				AddInDebugOutput(NULL
					, TEXT("InvokePropGetHelper")
					, IDS_DEBUG_IDISPATCHGETIDSOFNAMESFAILED
					, NULL); 
				MYRAISEEXCEPTION;
			}
		}

		return pDispatch->Invoke(*pdispid, IID_NULL
			, LOCALE_SYSTEM_DEFAULT, DISPATCH_PROPERTYGET
			, pDispParams, pvResult, NULL ,NULL);
	}

	EXCEPT{
   		AddInDebugOutput(NULL, TEXT("InvokePropGetHelper")
   			, IDS_DEBUG_EXCEPTION, NULL); 

		return ResultFromScode(E_FAIL);
	}
}

/*
 * InvokeMethodHelper
 *
 * Purpose:
 *  Wrapper for IDispatch->GetIDsOfNames and Invoke to help in calling
 *	Property Puts
 *
 * Parameters:
 *  pDispatch			IDispatch to use
 *  pdispid			DISPID if already known
 *  pDispParams		params to send
 *  pszMember			Name of Method to call
 *	pvResult			Result of call
 *
 * Return Value:
 *  HRESULT
 *
 */

HRESULT InvokePropPutHelper(LPDISPATCH pDispatch, DISPID FAR *pdispid, 
		DISPPARAMS FAR *pDispParams, LPTSTR pszMember, VARIANT FAR* pvResult){

	TRY{		
		if(FAILED(CheckDispatch(pDispatch))){
    		AddInDebugOutput(NULL, TEXT("CDataObject::Create")
    			, IDS_DEBUG_BADPOINTERTOXRTOBJECT, NULL); 
			return ResultFromScode(E_FAIL);
		}

		if(*pdispid == -1){
			HRESULT hr = pDispatch->GetIDsOfNames(IID_NULL,  (LPTSTR FAR *) &pszMember, 1
				, LOCALE_SYSTEM_DEFAULT, pdispid);
	
			if(FAILED(hr)){
				*pdispid = -1;
				AddInDebugOutput(NULL
					, TEXT("InvokePropPutHelper")
					, IDS_DEBUG_IDISPATCHGETIDSOFNAMESFAILED
					, NULL); 
				MYRAISEEXCEPTION;
			}
		}
			
		return pDispatch->Invoke(*pdispid, IID_NULL
			, LOCALE_SYSTEM_DEFAULT, DISPATCH_PROPERTYPUT
			, pDispParams, pvResult, NULL ,NULL);
	}
				

	EXCEPT{
   		AddInDebugOutput(NULL, TEXT("InvokePropPutHelper")
   			, IDS_DEBUG_EXCEPTION, NULL); 

		return ResultFromScode(E_FAIL);
	}
}

/*
 * VariantToXLOPER
 *
 * Purpose:
 *	Heper function which converts a VARIANT to an XLOPER
 *
 * Parameters:
 *  pxloper			XLOPER to receive data
 *  pvar				VARIANT hosting data
 *
 * Return Value:
 *  HRESULT
 *
 */

HRESULT VariantToXLOPER(LPXLOPER pxloper, VARIANT FAR *pvar){
	TRY{
		if(FAILED(CheckPointer((LPVOID) pvar, 0))){
    		AddInDebugOutput(NULL, TEXT("VariantToXLOPER")
    			, IDS_DEBUG_BADPOINTER, NULL); 
			return ResultFromScode(E_FAIL);
		}
		
		switch(pvar->vt){
			case VT_BSTR:
				return BSTRToXLOPER(pxloper, pvar->bstrVal);
			
			case VT_I2:
				pxloper->xltype = xltypeInt;
				pxloper->val.w = pvar->iVal;
				break;
			
			case VT_I4:
				pxloper->xltype = xltypeNum;
				pxloper->val.num = pvar->lVal;
				break;
						
			case VT_R4:
				pxloper->xltype = xltypeNum;
				pxloper->val.num = pvar->fltVal;
				break;
				
			case VT_R8:
				pxloper->xltype = xltypeNum;
				pxloper->val.num = pvar->dblVal;
				break;
			
			case VT_VARIANT | VT_BYREF:
				break;    		

    		case 0:
				pxloper->xltype = xltypeInt;
				pxloper->val.w = 0;
				break; 
			
			default:                                           
		   		AddInDebugOutput(NULL, TEXT("VariantToXLOPER")
   					, IDS_DEBUG_VTNOTSUPPORTED, NULL); 
				return ResultFromScode(E_FAIL);
		}
		return S_OK;   	
   	}                                  
   	
    #ifdef WIN32
	EXCEPT{
   		AddInDebugOutput(NULL, TEXT("VariantToXLOPER")
   			, IDS_DEBUG_EXCEPTION, NULL); 

		return ResultFromScode(E_FAIL);
	}           
	#endif
}

/*
 * BSTRToXLOPER
 *
 * Purpose:
 *	Heper function which converts a BSTR to an XLOPER.str
 *
 * Parameters:
 *  pxloper				XLOPER to receive data
 *  bstr				BSTR hosting data
 *  bFreeMalloc			If TRUE, IMalloc is released, other params 
 * 						should be NULL. Default = FALSE
 *
 * Return Value:
 *  HRESULT
 *
 */						 

HRESULT BSTRToXLOPER(LPXLOPER pxloper, BSTR bstr, BOOL bFreeMalloc){
	TRY{
		LPSTR psz = NULL;
		static LPMALLOC pMalloc = NULL;
		char ansibuf[MAX_STRING_SIZE + 4];
        
        if(bFreeMalloc){
			if(pMalloc){        	
        		pMalloc->Release();
        		pMalloc = NULL;
        	}
        		
        	return S_OK;
        }

		// XL supports strings of only 512 in len
		int len = min(lstrlen(bstr), MAX_STRING_SIZE);
        
        // I use IMalloc because...
        // 	SysAllocString returns WideChars in Win32 and XLNT is ANSI
        //  GlobalAllocPtr has limited selectors
        //	
        //	IMalloc is quick and is perfect for many small allocs

		// Get IMalloc and save it
		if(pMalloc == NULL){
			CoGetMalloc(MEMCTX_TASK, &pMalloc);
		}
			 
		// Alloc the string
		psz = (LPSTR) pMalloc->Alloc(MAX_STRING_SIZE + sizeof(DWORD));

		if(psz == NULL){
			AddInDebugOutput(NULL
				, TEXT("BSTRToXLOPER")
				, IDS_DEBUG_IMALLOCALLOCFAILED
				, NULL); 
		      	MYRAISEEXCEPTION;
		}

		memset(psz, 0, MAX_STRING_SIZE);
		memset(ansibuf, 0, sizeof(ansibuf));

		#ifdef UNICODE		
			WideCharToMultiByte(CP_ACP, NULL
				, bstr, len
				, ansibuf, MAX_STRING_SIZE, NULL, NULL);
		#else
			// XL accepts strings of only 256 chars. I truncate if its over.
			lstrcpyn(ansibuf, bstr, len + 1);	
		#endif

		
		// Create XLOPER strings which is len + string
		#ifdef _WIN32
			*psz = (char) len;
			lstrcatA(psz, (LPSTR) ansibuf);
        #else
			int nStrLen = lstrlen(ansibuf);
			*psz = (char) len;
			lstrcat(psz, (LPSTR) ansibuf);
        #endif
        
		// Save xloper
		pxloper->val.str = psz;
		pxloper->xltype = xltypeStr;
		
		if(pxloper->val.str){	
			return S_OK;   	
		}else{
			return ResultFromScode(E_FAIL);   	
		}
   	}

	EXCEPT{
   		AddInDebugOutput(NULL, TEXT("BSTRToXLOPER")
   			, IDS_DEBUG_EXCEPTION, NULL); 

		return ResultFromScode(E_FAIL);   	
	}
}

/*
 * MyXLSet
 *
 * Purpose:
 *	Helper function which eases xlSet. 
 *	If SheetID is cached and FileName and SheetName == NULL then
 * 	I don't try to get the sheetid again to optimize.
 *
 *
 * Parameters:
 *  pCache				Pointer to DataObject's cache
 *  bstrFileName		FileName containing worksheet top update
 *	bstrSheetName		Worksheet to update
 *  uiRowStart			The row of upper-left cell to place update on sheet
 *  ucColStart			The col of upper-left cell to place update on sheet
 *	uiWidth				The width of range to update
 *	uiHeight			The height of range to update	
 *
 * Return Value:
 *  HRESULT
 *
 */

HRESULT MyXLSet(LPXLOPER pxlDataObject, BSTR bstrFileName
		, BSTR bstrSheetName, UINT uiRowStart
		, BYTE ucColStart, UINT uiWidth, UINT ucHeight){

    XLMREF xlmref;						    
	XLOPER xArray, xRef;
	char ansibuf[128];
	int len;

	static XLOPER xSheetID;
	static BOOL bGotSheetID = FALSE;

	TRY{	
		if(	bstrFileName == NULL 
				&& bstrSheetName == NULL
				&& bGotSheetID == FALSE){

	   		AddInDebugOutput(NULL, TEXT("MyXLSet")
   				, IDS_DEBUG_XLSHEEDIDFAILED, NULL); 
    	
    		return ResultFromScode(E_FAIL);
		}

		if(	(bstrFileName != NULL && bstrSheetName != NULL)
				|| bGotSheetID == FALSE){

			// First we need to get the SheetID
			xSheetID.xltype = xltypeStr;
		     
			#ifdef UNICODE                          
				// Excel NT is ANSI
				char szFileName[128];
				char szSheetName[128];

				memset(szFileName, 0, 128);
				memset(szSheetName, 0, 128);
	
				len = SysStringLen(bstrFileName);

				WideCharToMultiByte(CP_ACP, NULL
					, bstrFileName ,len
					, szFileName, 128, NULL, NULL);
	
				len = SysStringLen(bstrSheetName);

				WideCharToMultiByte(CP_ACP, NULL
					, bstrSheetName ,len
					, szSheetName, 128, NULL, NULL);
			
				len = lstrlenA(szFileName) + lstrlenA(szSheetName) + (2 * sizeof(char));
	
				wsprintfA(ansibuf, "%c[%s]%s", len
					, (LPSTR) szFileName, (LPSTR) szSheetName);
			#else
				len = lstrlen(bstrFileName) + lstrlen(bstrSheetName) + 2;
			
				wsprintf(ansibuf, "%c[%s]%s", len
				, (LPSTR) bstrFileName, (LPSTR) bstrSheetName);
			#endif

			xSheetID.val.str = ansibuf;
	
    		if(xlretSuccess != Excel(xlSheetId, (LPXLOPER) &xRef, 1, (LPXLOPER) &xSheetID)){
	   			AddInDebugOutput(NULL, TEXT("MyXLSet")
   				, IDS_DEBUG_XLSHEEDIDFAILED, NULL); 
    			return ResultFromScode(E_FAIL);
    		}

			bGotSheetID = TRUE;
		}
	    
	    // Now, do the  xlSet
	    
	    xRef.xltype = xltypeRef;
    	xRef.val.mref.lpmref = (LPXLMREF) &xlmref;
	
		xlmref.count = 1;
		xArray.xltype = xltypeMulti;               
	
   		xlmref.reftbl[0].rwFirst = uiRowStart;   
		xlmref.reftbl[0].rwLast = (uiRowStart + ucHeight - 1);   
		xlmref.reftbl[0].colFirst = ucColStart;   
		xlmref.reftbl[0].colLast = (ucColStart + uiWidth - 1);   

		xArray.val.array.lparray = (LPXLOPER) pxlDataObject;
		xArray.val.array.rows = ucHeight;
		xArray.val.array.columns = uiWidth;

	   	if(xlretSuccess != Excel(xlSet, 0, 2, (LPXLOPER) &xRef, (LPXLOPER) &xArray)){
	   		AddInDebugOutput(NULL, TEXT("MyXLSet")
   				, IDS_DEBUG_XLSETFAILED, NULL); 
   			return ResultFromScode(E_FAIL);
	   	}

	   	return S_OK;
	}

	#ifdef _WIN32	    
	EXCEPT{
   		AddInDebugOutput(NULL, TEXT("MyXLSet")
   			, IDS_DEBUG_EXCEPTION, NULL); 

		return ResultFromScode(E_FAIL);
	}
	#endif
}


/*
 * XRTLoadPersistFile
 *
 * Purpose:
 *	Calls object's IPersistFile::Load and returns object's IDispatch
 *
 * Parameters:
 *  pszFileName			FileName of object's storage to load
 *	ppDispatch			Pointer to LPDISPATCH to return
 *
 * Return Value:
 *  HRESULT
 *
 */


HRESULT XRTLoadPersistFile(LPTSTR pszFileName, LPDISPATCH FAR *ppDispatch){
	LPPERSISTFILE pXRTPersistFile = NULL;
	LPDISPATCH pXRTDispatch = NULL;
	CLSID clsid;

	TRY{
		if(FAILED(CheckPointer((LPVOID) ppDispatch, 0))){
    		AddInDebugOutput(NULL, TEXT("XRTLoadPersistFile")
    			, IDS_DEBUG_BADPOINTER, NULL); 
			return ResultFromScode(E_FAIL);
		}
	    
	    *ppDispatch = NULL;
	
		if(FAILED(GetClassFile(pszFileName, &clsid))){
			AddInDebugOutput(NULL
				, TEXT("XRTLoadPersistFile")
				, IDS_DEBUG_GETCLASSFILEFAILED
				, DISP_E_EXCEPTION); 
			MYRAISEEXCEPTION;
		}	                   
	
		if(FAILED(CoCreateInstance(clsid, NULL, CLSCTX_SERVER, IID_IPersistFile
				, (LPVOID FAR *)  &pXRTPersistFile))){ 
			AddInDebugOutput(NULL
				, TEXT("XRTLoadPersistFile")
				, IDS_DEBUG_COCREATEINSTANCEFAILED
				, DISP_E_EXCEPTION); 
			MYRAISEEXCEPTION;
		}	                   
                  
		if(FAILED(pXRTPersistFile->Load(pszFileName
				, STGM_DIRECT | STGM_SHARE_DENY_NONE | STGM_READ))){
			AddInDebugOutput(NULL
				, TEXT("XRTLoadPersistFile")
				, IDS_DEBUG_IPERSISTFILELOADFAILED
				, DISP_E_EXCEPTION); 
			MYRAISEEXCEPTION;
    	}
	        
		if(FAILED(pXRTPersistFile->QueryInterface(IID_IDispatch
				, (LPVOID FAR *) ppDispatch))){
			AddInDebugOutput(NULL
				, TEXT("XRTLoadPersistFile")
				, IDS_DEBUG_QIFORIDISPATCHFAILED
				, DISP_E_EXCEPTION); 
			MYRAISEEXCEPTION;
	    }
    
		pXRTPersistFile->Release();
		pXRTPersistFile = NULL;
	
		return S_OK;
	}

	EXCEPT{	
		if(pXRTPersistFile) 
			pXRTPersistFile->Release;
	
		if(*ppDispatch) 
			(*ppDispatch)->Release;
	
   		AddInDebugOutput(NULL, TEXT("XRTLoadPersistFile")
   			, IDS_DEBUG_EXCEPTION, NULL); 

		return ResultFromScode(E_FAIL);
	}
}


HRESULT XLRunMacro(CDataObject FAR *pDataObject){
	static DISPID dispid = -1;
	static TCHAR szMember[] = TEXT("Run");
	static DISPPARAMS dispparams;
	VARIANT param;

	TRY{		
		VariantInit(&param);
	
		param.vt = VT_BSTR;
		param.bstrVal = SysAllocString(pDataObject->m_bstrXLCallBackFunc);
	
		dispparams.rgvarg = &param;
		dispparams.rgdispidNamedArgs = NULL;
		dispparams.cArgs = 1;
		dispparams.cNamedArgs = 0;

		if(FAILED(InvokeMethodHelper(g_pRealTimeExt->m_pXLAppDispatch, &dispid, 
				&dispparams, szMember, NULL))){
			AddInDebugOutput(NULL
				, TEXT("XLRunMacro")
				, IDS_DEBUG_QIFORIDISPATCHFAILED
				, DISP_E_EXCEPTION); 
			MYRAISEEXCEPTION;
		}
			
		VariantClear(&param);
	
		return S_OK;
	}

	EXCEPT{
		return ResultFromScode(E_FAIL);
	}
}

HRESULT XLCalculateNow(CDataObject FAR *pDataObject){
	static DISPID dispid = -1;
	static DISPPARAMS dispparams = {NULL, NULL, 0, 0};
	static TCHAR szMember[] = TEXT("Calculate");
	HRESULT hr;
	VARIANT vResult;

	VariantInit(&vResult);
	
	hr = InvokeMethodHelper(pDataObject->m_pUpdateRangeDispatch, &dispid, 
		&dispparams, szMember, &vResult);

	VariantClear(&vResult);
	
	return hr;		
}

HRESULT CheckDispatch(LPDISPATCH lpDispatch){
	TRY{
		if(FAILED(CheckPointer((LPVOID) lpDispatch, 0))){
			return ResultFromScode(E_FAIL);
		}

		if(FAILED((lpDispatch)->QueryInterface(IID_IDispatch
				, (LPVOID FAR*) &lpDispatch))){
			return ResultFromScode(E_FAIL);
		}

		// Release from QI
    	lpDispatch->Release();
    	lpDispatch = NULL;

		return S_OK;
	}		

    #ifdef WIN32
	EXCEPT{
		return ResultFromScode(E_FAIL);
	}
	#endif
}

HRESULT CheckVariant(VARIANT FAR *pv, BOOL bDeref){
	TRY{
		if(FAILED(CheckPointer((LPVOID) pv, 0))){
			return ResultFromScode(E_FAIL);
		}

		if(bDeref && pv->vt & VT_BYREF){
			VariantCopyInd(pv, pv);
		}

		if(pv->vt == VT_EMPTY || pv->vt == VT_ERROR){
			return ResultFromScode(E_FAIL);
		}

		if(pv->vt == VT_BSTR){
			if(pv->bstrVal == NULL){
				return ResultFromScode(E_FAIL);
			}

			if(FAILED(CheckPointer((LPVOID) pv->bstrVal, 0))){
				return ResultFromScode(E_FAIL);
			}
		}else if(pv->vt == VT_DISPATCH){
			if(pv->pdispVal == NULL){
				return ResultFromScode(E_FAIL);
			}

			if(FAILED(CheckDispatch(pv->pdispVal))){
				return ResultFromScode(E_FAIL);
			}
		}else if(pv->vt == VT_VARIANT){
			if(FAILED(CheckVariant(pv, bDeref))){
				return ResultFromScode(E_FAIL);
			}
		}

		return S_OK;
	}		

    #ifdef WIN32
	EXCEPT{
		return ResultFromScode(E_FAIL);
	}
	#endif
}


void AddInDebugOutput(LPEXCEPINFO pexcepinfo, LPTSTR pszFunction
		, UINT wStrID, SCODE scode)
{
	TCHAR buf[MAX_DEBUG_STRING_LEN]; // These are internal strings. I know len < 256

	#ifdef WIN32
		HANDLE hMod = GetModuleHandle(FILENAME);
	#else
		HANDLE hMod = g_hinst;	
	#endif

	if(NULL == LoadString(hMod, wStrID, buf, sizeof(buf))){
		lstrcpy(buf, TEXT("Internal error. Cannot provide desc of error."));
	}
	
	if((pexcepinfo && wStrID != IDS_DEBUG_EXCEPTION )
		|| (pexcepinfo && wStrID == IDS_DEBUG_EXCEPTION && (!pexcepinfo->wCode)) ){
		
		pexcepinfo->scode = NULL;
		pexcepinfo->wCode = (UINT) ERROR_CODE_OFFSET + wStrID;
		pexcepinfo->bstrDescription = SysAllocString(buf);
	}

	TCHAR szDebugOutput[MAX_DEBUG_STRING_LEN]; // These are internal strings. I know len

	lstrcpy(szDebugOutput, pszFunction);
	lstrcat(szDebugOutput, TEXT(" : "));
	lstrcat(szDebugOutput, buf);
	lstrcat(szDebugOutput, TEXT("\r\n"));

	#ifdef USEDEBUGWINDOW
 		OutputDebugString((LPTSTR) szDebugOutput);
    #else
 		MessageBox(g_hwndXL, (LPTSTR) szDebugOutput, (LPTSTR) g_szTitle, MB_OK);
 	#endif
}

void AddInDebugOutput(LPEXCEPINFO pexcepinfo, LPTSTR pszFunction
		, LPTSTR pszDesc, SCODE scode)
{
	if((pexcepinfo && scode != IDS_DEBUG_EXCEPTION )
		|| (pexcepinfo && scode == IDS_DEBUG_EXCEPTION && (!pexcepinfo->scode || !pexcepinfo->wCode)) ){
		
		if(scode && FALSE){
			pexcepinfo->scode = scode;
			pexcepinfo->wCode = 0;
			pexcepinfo->bstrDescription = NULL;
		}else{
			pexcepinfo->scode = NULL;
			pexcepinfo->wCode = (UINT) CACHE_ERROR_CODE;
			pexcepinfo->bstrDescription = SysAllocString(pszDesc);
		}
	}

	TCHAR szDebugOutput[MAX_DEBUG_STRING_LEN]; // These are internal strings. I know len

	lstrcpy(szDebugOutput, pszFunction);
	lstrcat(szDebugOutput, TEXT(" : "));
	lstrcat(szDebugOutput, pszDesc);
	lstrcat(szDebugOutput, TEXT("\r\n"));
	
	#ifdef USEDEBUGWINDOW
 		OutputDebugString((LPTSTR) szDebugOutput);
    #else
 		MessageBox(g_hwndXL, (LPTSTR) szDebugOutput, (LPTSTR) g_szTitle, MB_OK);
 	#endif
}

HRESULT CheckPointer(LPVOID pVoid, UINT uiSize){
	TRY{
		if(pVoid == NULL){
			return ResultFromScode(E_FAIL);
        }
            
        if(FALSE == IsValidAddress(pVoid, uiSize, TRUE)){
			return ResultFromScode(E_FAIL);
        }
            
		return S_OK;
	}

	#ifdef WIN32	
	EXCEPT{
		return ResultFromScode(E_FAIL);
	}           
	#endif
}


#ifdef WIN32
BOOL IsValidAddress(void* lp, UINT nBytes, BOOL bReadWrite)
{
	// simple version using Win-32 APIs for pointer validation.
	return (lp != NULL && !IsBadReadPtr(lp, nBytes) &&
		(!bReadWrite || !IsBadWritePtr((LPVOID)lp, nBytes)));
}
#else
BOOL IsValidAddress(void FAR* lp, UINT nBytes, BOOL bReadWrite)
{
	if(IsBadWritePtr(lp, nBytes))
		return FALSE;

	if(IsBadReadPtr(lp, nBytes))
		return FALSE;

	return TRUE;
}
#endif

void PrepareExcepInfo(EXCEPINFO FAR *pexcepinfo){
   	pexcepinfo->pvReserved = NULL;
	pexcepinfo->bstrHelpFile = NULL; 
	pexcepinfo->scode = 0;
	pexcepinfo->pfnDeferredFillIn = NULL;
	pexcepinfo->dwHelpContext = 0;

	return;
}

