// 	WOSA/XRT Excel 5.0 AddIn Version 1.00
//
//	COPYRIGHT:
//
//  	(C) Copyright Microsoft Corp. 1994.  All rights reserved.
//
//   	You have a royalty-free right to use, modify, reproduce and
//   	distribute the Sample Files (and/or any modified version) in
//   	any way you find useful, provided that you agree that
//   	Microsoft has no warranty obligations or liability for any
//   	Sample Application Files which are modified.
//
//
// 	Revision History
//    	October 15, 1994   rc    Version 1.00 Release
//
// 	Comments:
// 		The add-ins CProeprty really only returns the proeprty name. Remeber,
// 		the DataItem(s) hierarchy is totaly implemented by the bound WOSA/XRT object
//		 and proeprty spawning from a request should provide a proeprties name only.


#include "precomp.h"
#include "guid.h"

#include "prop.h"
#include "props.h"
          
static HRESULT hr;

/*
 * CProperty::Create
 *
 * Purpose:
 *  Creates an instance of a CProperty and initializes it.
 *
 * Parameters:
 *	bstrName		Name of the property
 *	pXRTDispatch	Property's respective XRT property
 *  ppProperty    	Pointer to return created CProperty.
 *
 * Return Value:
 *  HRESULT
 *
 */ 
 
HRESULT CProperty::Create(BSTR bstrName, LPDISPATCH pXRTDispatch
		, CProperty FAR* FAR* ppProperty)
{   
    CProperty FAR* pProperty;

	TRY{
		if(FAILED(CheckPointer((LPVOID) ppProperty, 0))){
    		AddInDebugOutput(NULL, TEXT("CProperty::Create")
    			, IDS_DEBUG_BADPOINTER, NULL); 
			return ResultFromScode(E_FAIL);
		}
	
    	*ppProperty = NULL;

		// In create properties I will not get a dispatch for Name
		// nor do I need one.
		if(SUCCEEDED(CheckDispatch(pXRTDispatch))){
			pProperty->m_pXRTDispatch = pXRTDispatch;
		}

    	pProperty = new CProperty();
    
    	if(pProperty == NULL){
			AddInDebugOutput(&pProperty->m_excepinfo
					, TEXT("CProperty::Create")
					, IDS_DEBUG_CREATEPROPERTYFAILED
					, NULL); 
			MYRAISEEXCEPTION;
	    }
     
    
    	// Load type information for the collection from type library. CreateStdDispatch
    	// is not used to create a standard IDispatch implemenation because this
    	// collection implementation raises exceptions on ResultFromScode(E_FAIL)s.    
	    if (FAILED(LoadTypeInfo(&pProperty->m_ptinfo, IID_IProperty))){
			AddInDebugOutput(&pProperty->m_excepinfo
					, TEXT("CProperty::Create")
					, IDS_DEBUG_COULDNOTLOADTYPEINFO
					, NULL); 
			MYRAISEEXCEPTION;
	    }

		pProperty->m_bstrName = SysAllocString(bstrName);                        
    		    
    	*ppProperty = pProperty;       
    
    	return S_OK;
	}
    
	EXCEPT{
    	if (pProperty == NULL)
        	return ResultFromScode(E_FAIL);
	
   		AddInDebugOutput(NULL, TEXT("CProperty::Create")
   			, IDS_DEBUG_EXCEPTION, NULL); 
        	
    	if(pProperty)
    		delete pProperty;

	    return ResultFromScode(E_FAIL);
	}
}

/*
 * CProperty::CProperty
 *
 * Purpose:
 *  Constructor for CProperty object.
 *
 */                             
 
CProperty::CProperty()
{
	m_pXRTDispatch = NULL;
	m_pProperties = NULL;
	m_pRequest = NULL;
	m_bDirectUpdates = FALSE;
	m_bstrName = NULL;
	m_ptinfo = NULL;
	m_cRef = 1;
   	m_bRaiseException = 0;
   	m_pxlProperty = NULL;
}

/*
 * CProperty::~CProperty
 *
 * Purpose:
 *  Destructor for CProperty object. 
 *
 */

CProperty::~CProperty()
{                                
	if(SUCCEEDED(CheckPointer(m_bstrName)))
		SysFreeString(m_bstrName);

	if(SUCCEEDED(CheckPointer(m_ptinfo)))
		m_ptinfo->Release();

	if(SUCCEEDED(CheckDispatch(m_pXRTDispatch)))
   		m_pXRTDispatch->Release();
   		
   	if(SUCCEEDED(CheckPointer(m_pxlProperty)) 
   			&& m_pxlProperty->xltype == xltypeStr 
   			&& m_pxlProperty->val.str){

		LPMALLOC pMalloc = NULL;

		CoGetMalloc(MEMCTX_TASK, &pMalloc);
   		pMalloc->Free(m_pxlProperty->val.str);
		pMalloc->Release();
   	}
}


/*
 * CProperty::QueryInterface, AddRef, Release
 *
 * Purpose:
 *  Implements IUnknown::QueryInterface, AddRef, Release
 *
 */
 
STDMETHODIMP
CProperty::QueryInterface(REFIID iid, void FAR* FAR* ppv) 
{   
    *ppv = NULL;
	
    if (iid == IID_IUnknown) 
		*ppv = this;  
    else if (iid == IID_IProperty) 
		*ppv = this;    
    else if (iid == IID_IDispatch || iid == IID_DProperty)
		*ppv = this;   
    else 
    	return ResultFromScode(E_NOINTERFACE); 

    AddRef();
    return S_OK;    
}


STDMETHODIMP_(ULONG)
CProperty::AddRef(void)
{ 
    return ++m_cRef;
}


STDMETHODIMP_(ULONG)
CProperty::Release(void)
{ 
    if(--m_cRef == 0){
		delete this;
		return 0;
    }
    return m_cRef;
} 


/*
 * CProperty::GetTypeInfoCount
 *
 * Purpose:
 *  Implements IDispatch::GetTypeInfoCount.
 *
 */
STDMETHODIMP
CProperty::GetTypeInfoCount(UINT FAR* pctinfo)
{
    *pctinfo = 1;
    return S_OK;
}

/*
 * CProperty::GetTypeInfo
 *
 * Purpose:
 *  Implements IDispatch::GetTypeInfo. 
 *
 */
STDMETHODIMP
CProperty::GetTypeInfo(
      UINT itinfo,
      LCID lcid,
      ITypeInfo FAR* FAR* pptinfo)
{    
    *pptinfo = NULL;
     
    if(itinfo != 0)
		return ResultFromScode(DISP_E_BADINDEX);
    
    m_ptinfo->AddRef(); 
    *pptinfo = m_ptinfo;
    
    return S_OK;
}

/*
 * CProperty::GetIDsOfNames
 *
 * Purpose:
 *  Implements IDispatch::GetIDsOfNames.  The standard implementation, DispGetIDsOfNames,
 *  is used.
 *
 */
STDMETHODIMP 
CProperty::GetIDsOfNames(
      REFIID riid,
      OLECHAR FAR* FAR* rgszNames,
      UINT cNames,
      LCID lcid,
      DISPID FAR* rgdispid)
{
    return DispGetIDsOfNames(m_ptinfo, rgszNames, cNames, rgdispid);
}

/*
 * CProperty::Invoke
 *
 * Purpose:
 *  Implements IDispatch::Invoke.  The standard implementation, DispInvoke,
 *  is used. Properties and methods exposed by the collection object will
 *  set m_bRaiseException to raise an exception. 
 *
 *	In Excel RealTime Extensions version 1.00, I could not get automation
 *	exception exceptions to occurr consistently with descriptions and 
 *  wCodes. 
 *
 */
STDMETHODIMP
CProperty::Invoke(
      DISPID dispidMember,
      REFIID riid,
      LCID lcid,
      WORD wFlags,
      DISPPARAMS FAR* pdispparams,
      VARIANT FAR* pvarResult,
      EXCEPINFO FAR* pexcepinfo,
      UINT FAR* puArgErr)
{  
    HRESULT hr = S_OK;
     
    if (NULL == pexcepinfo)
		return ResultFromScode(E_INVALIDARG); 

    m_bRaiseException = FALSE;
	 
    hr = DispInvoke(          
		this, m_ptinfo,
		dispidMember, wFlags, pdispparams,
		pvarResult, pexcepinfo, puArgErr);
    
    // Check if property accessor function or method called by DispInvoke wants to raise
    // an exception. If so, return DISP_E_EXCEPTION after filling out the EXCEPINFO 
    // structure. The scode and bstrDescription fields of EXCEPINFO are filled by the 
    // property or method.    
    if (m_bRaiseException){
		pexcepinfo->bstrDescription = m_excepinfo.bstrDescription;
		pexcepinfo->bstrSource = SysAllocString(g_szTitle);   
		pexcepinfo->wCode = m_excepinfo.wCode;   

		PrepareExcepInfo(pexcepinfo);

		m_bRaiseException = FALSE;
        m_excepinfo.scode = NULL;
        m_excepinfo.wCode = 0;

		return ResultFromScode(DISP_E_EXCEPTION); 
    }else{
    	return hr;    
	}
}


/*
 * CProperty::Name - Read-only property
 *
 * Purpose:
 *	Returns Name of property
 *
 * Return Value:
 *  BSTR
 *
 */

 
STDMETHODIMP_(BSTR)     
CProperty::get_Name(void)  
{   
	TRY{     
		return SysAllocString(m_bstrName);		
	}

	#ifdef _WIN32
	EXCEPT{
   		AddInDebugOutput(NULL, TEXT("CProperty::Name")
   			, IDS_DEBUG_EXCEPTION, NULL); 

		m_bRaiseException = TRUE;

		return NULL;
	}
	#endif
}

/*
 * CProperty::Value - Read-only property
 *
 * Purpose:
 *	Returns Value of property
 *
 * Return Value:
 *  VARIANT
 *
 */
 
STDMETHODIMP_(VARIANT)     
CProperty::get_Value(void)  
{
	// Since we only implemet property for request(s) the value should retutn
	// an error. DataItem(s) is passed on to the WOSAXRT object.
    VARIANT v;
    
    VariantInit(&v);
    
    v.vt = VT_ERROR;
           
	return v;
}

/*
 * CProperty::Object - Read-only property
 *
 * Purpose:
 *	Returns WOSA/XRT object CProperty object. Used by VBA developer to call
 *	properties and methods which are specific to the WOSA/XRT object.
 *
 * Return Value:
 *  VARIANT
 *
 */

STDMETHODIMP_(IDispatch *)
CProperty::get_Object(){
	if(FAILED(CheckPointer(m_pXRTDispatch, 0))){
		AddInDebugOutput(&m_excepinfo
				, TEXT("CProperty::get_Object")
				, IDS_DEBUG_BADPOINTER
				, DISP_E_EXCEPTION); 

		m_bRaiseException = TRUE;

		return NULL;
	}

	m_pXRTDispatch->AddRef();	
                   
	return m_pXRTDispatch;                     
}


