// 	WOSA/XRT Excel 5.0 AddIn Version 1.00
//
//	COPYRIGHT:
//
//  	(C) Copyright Microsoft Corp. 1994.  All rights reserved.
//
//   	You have a royalty-free right to use, modify, reproduce and
//   	distribute the Sample Files (and/or any modified version) in
//   	any way you find useful, provided that you agree that
//   	Microsoft has no warranty obligations or liability for any
//   	Sample Application Files which are modified.
//
//
// 	Revision History
//    	October 15, 1994   rc    Version 1.00 Release
//
// 	Comments:
// 		The CProperties class implements the IUnknown and IDispatch interfaces.
// 		The IDispatch interface supersets all the WOSA/XRT Properties methods 
// 		and properties. The other properties and methods are unique to to add-in
// 		and are defined throughout this module... 

#include "precomp.h"
#include "guid.h"

#include "dataobj.h"
#include "requests.h"
#include "request.h"
#include "props.h"
#include "prop.h"
#include "enumvar.h"
#include "cache.h"

static HRESULT hr;
          
/*
 * CProperties::Create
 *
 * Purpose:
 *  Creates an instance of a collection object and initializes it.
 *
 * Return Value:
 *  HRESULT
 *
 */

HRESULT CProperties::Create(CProperties FAR* FAR* ppProperties) 
{   
    CProperties FAR* pProperties = NULL;

	TRY{     
		if(FAILED(CheckPointer((LPVOID) ppProperties, 0))){
    		AddInDebugOutput(NULL, TEXT("CProperties::Create")
    			, IDS_DEBUG_BADPOINTER, NULL); 
			return ResultFromScode(E_FAIL);
		}
	
    	*ppProperties = NULL;
    
    	// Create new collection
    	pProperties = new CProperties();
	    if (pProperties == NULL){
			AddInDebugOutput(&pProperties->m_excepinfo
					, TEXT("CProperties::Create")
					, IDS_DEBUG_PROPERTIESALLOCFAILED
					, NULL); 
			MYRAISEEXCEPTION;
	    }
    
	    // Load type information for the collection from type library. CreateStdDispatch
    	// is not used to create a standard IDispatch implemenation because this
    	// collection implementation raises exceptions on ResultFromScode(E_FAIL)s.    
	    if (FAILED(LoadTypeInfo(&pProperties->m_ptinfo, IID_IProperties))){
			AddInDebugOutput(&pProperties->m_excepinfo
					, TEXT("CProperties::Create")
					, IDS_DEBUG_COULDNOTLOADTYPEINFO
					, NULL); 
			MYRAISEEXCEPTION;
    	}

	    *ppProperties = pProperties;

    	return S_OK;
	}
    
	EXCEPT{
    	if (pProperties == NULL)
			return ResultFromScode(E_FAIL);

   		AddInDebugOutput(NULL, TEXT("CProperties::Create")
   			, IDS_DEBUG_EXCEPTION, NULL); 

	    delete pProperties;

	    return ResultFromScode(E_FAIL);
	}
}

/*
 * CProperties::CProperties
 *
 * Purpose:
 *  Constructor for CProperties object. Initializes members to NULL.
 *
 */

CProperties::CProperties()
{    
	m_pDataObject = NULL;
	m_pRequest = NULL;    
	m_pXRTDispatch = NULL;
	m_bstrName = NULL;
	m_ptinfo = NULL;
	m_cRef = 1; 
   	m_bRaiseException = 0;
}

/*
 * CProperties::~CProperties
 *
 * Purpose:
 *  Destructor for CProperties object. 
 *
 */

CProperties::~CProperties()
{    
	if(SUCCEEDED(CheckDispatch(m_pXRTDispatch)))
   		m_pXRTDispatch->Release();

	if(SUCCEEDED(CheckPointer(m_bstrName)))
		SysFreeString(m_bstrName);

	if(SUCCEEDED(CheckPointer(m_ptinfo)))
		m_ptinfo->Release();	
}


/*
 * CProperties::QueryInterface, AddRef, Release
 *
 * Purpose:
 *  Implements IUnknown::QueryInterface, AddRef, Release
 *
 */
 
STDMETHODIMP
CProperties::QueryInterface(REFIID iid, void FAR* FAR* ppv) 
{   
    *ppv = NULL;
	
    if (iid == IID_IUnknown) 
		*ppv = this;  
    else if (iid == IID_IProperties) 
		*ppv = this;    
    else if (iid == IID_IDispatch || iid == IID_DProperties)
		*ppv = this;   
    else 
    	return ResultFromScode(E_NOINTERFACE); 

    AddRef();
    return S_OK;    
}


STDMETHODIMP_(ULONG)
CProperties::AddRef(void)
{ 
    return ++m_cRef;
}


STDMETHODIMP_(ULONG)
CProperties::Release(void)
{ 
    if(--m_cRef == 0){
		delete this;
		return 0;
    }
    return m_cRef;
} 


/*
 * CProperties::GetTypeInfoCount
 *
 * Purpose:
 *  Implements IDispatch::GetTypeInfoCount.
 *
 */

STDMETHODIMP
CProperties::GetTypeInfoCount(UINT FAR* pctinfo)
{
    *pctinfo = 1;
    return S_OK;
}

/*
 * CProperties::GetTypeInfo
 *
 * Purpose:
 *  Implements IDispatch::GetTypeInfo. 
 *
 */

STDMETHODIMP
CProperties::GetTypeInfo(
      UINT itinfo,
      LCID lcid,
      ITypeInfo FAR* FAR* pptinfo)
{    
    *pptinfo = NULL;
     
    if(itinfo != 0)
		return ResultFromScode(DISP_E_BADINDEX);
    
    m_ptinfo->AddRef(); 
    *pptinfo = m_ptinfo;
    
    return S_OK;
}

/*
 * CProperties::GetIDsOfNames
 *
 * Purpose:
 *  Implements IDispatch::GetIDsOfNames.  The standard implementation, DispGetIDsOfNames,
 *  is used.
 *
 */

STDMETHODIMP 
CProperties::GetIDsOfNames(
      REFIID riid,
      OLECHAR FAR* FAR* rgszNames,
      UINT cNames,
      LCID lcid,
      DISPID FAR* rgdispid)
{
    return DispGetIDsOfNames(m_ptinfo, rgszNames, cNames, rgdispid);
}

/*
 * CProperties::Invoke
 *
 * Purpose:
 *  Implements IDispatch::Invoke.  The standard implementation, DispInvoke,
 *  is used. Properties and methods exposed by the collection object will
 *  set m_bRaiseException to raise an exception. 
 *
 *	In Excel RealTime Extensions version 1.00, I could not get automation
 *	exception exceptions to occurr consistently with descriptions and 
 *  wCodes. 
 *
 */

STDMETHODIMP
CProperties::Invoke(
      DISPID dispidMember,
      REFIID riid,
      LCID lcid,
      WORD wFlags,
      DISPPARAMS FAR* pdispparams,
      VARIANT FAR* pvarResult,
      EXCEPINFO FAR* pexcepinfo,
      UINT FAR* puArgErr)
{  
    HRESULT hr = S_OK;
     
    if (NULL == pexcepinfo)
		return ResultFromScode(E_INVALIDARG); 

    m_bRaiseException = FALSE;
	 
    hr = DispInvoke(          
		this, m_ptinfo,
		dispidMember, wFlags, pdispparams,
		pvarResult, pexcepinfo, puArgErr);
    
    // Check if property accessor function or method called by DispInvoke wants to raise
    // an exception. If so, return DISP_E_EXCEPTION after filling out the EXCEPINFO 
    // structure. The scode and bstrDescription fields of EXCEPINFO are filled by the 
    // property or method.    
    if (m_bRaiseException){
		pexcepinfo->bstrDescription = m_excepinfo.bstrDescription;
		pexcepinfo->bstrSource = SysAllocString(g_szTitle);   
		pexcepinfo->wCode = m_excepinfo.wCode;   

		PrepareExcepInfo(pexcepinfo);

		m_bRaiseException = FALSE;
        m_excepinfo.scode = NULL;
        m_excepinfo.wCode = 0;

		return ResultFromScode(DISP_E_EXCEPTION); 
    }else{
    	return hr;    
	}
}


/*
 * CProperties::AddExistingProps
 *
 * Purpose:
 *  Adds existing propeties collection to another
 *
 * Parameters:
 *  pSrcProps	Properties to copy into this collection
 *
 * Return Value:
 *  HRESULT
 *
 */
 
HRESULT CProperties::AddExistingProps(CProperties FAR *pSrcProps) 
{   
	VARIANT v;
	DWORD i = 0;
	CNode FAR* pNode = NULL;
	CProperty FAR* pSrcProperty = NULL;
	CProperty FAR* pProperty = NULL;       

	TRY{    
		VariantInit(&v);
		
		// Add each property to the new Properties object
		for(i=0, pNode = pSrcProps->m_col.m_pFirstNode
				; i<pSrcProps->m_col.m_cElements && pNode
				; i++, pNode = pNode->m_pNext){
                
			// Cast as CProperty 
			// Internal QI just is a recast
			pNode->m_pDispatch->QueryInterface(IID_IProperty
					, (LPVOID FAR *) &pSrcProperty);
                
			// Add the property
			v.vt = VT_BSTR;
			v.bstrVal = SysAllocString(pSrcProperty->m_bstrName);

			// Add
        	if(FAILED(Add(v, &pProperty, 0L))){
				AddInDebugOutput(&m_excepinfo
						, TEXT("CProperties::Create")
						, IDS_DEBUG_PROPSADDFAILED
						, NULL); 
				MYRAISEEXCEPTION;
        	}

			// Yes we need to release here. We have no intention of 
			// keeping pProperty but Add AddRefed it for us. Maybe this 
			// overloaded Add should never addref but it does so I release.
			pProperty->Release();
			pProperty = NULL;
			
            pSrcProperty->Release();
            pSrcProperty = NULL;
            
	       	VariantClear(&v);
		}

	    return S_OK;
	}
	
	EXCEPT{    
		if(pNode && pNode->m_pDispatch)
			pNode->m_pDispatch->Release();

		if(pSrcProperty)
			pSrcProperty->Release();

		if(pProperty)
			pProperty->Release();			

       	VariantClear(&v);

		m_bRaiseException = TRUE;

   		AddInDebugOutput(&m_excepinfo, TEXT("CProperties::AddExistingProps")
   			, IDS_DEBUG_EXCEPTION, NULL); 

		return ResultFromScode(E_FAIL);
	}
}


/*
 * CProperties::Add
 *
 * Purpose:
 *  Automation Add - Creates and adds Cproperties to collection
 *
 * Parameters:
 *  pSrcProps	Properties to copy into this collection
 *
 * Return Value:
 *  HRESULT
 *
 */
 

STDMETHODIMP_(IDispatch *) 
CProperties::Add(VARIANT Name)
{
    LPDISPATCH pMyDispatch = NULL;
    LPDISPATCH pXRTDispatch = NULL;
    CProperty FAR* pProperty = NULL;

	TRY{
		if(FAILED(CheckVariant(&Name, TRUE))){
			AddInDebugOutput(&m_excepinfo
					, TEXT("CProperties::Add")
					, IDS_DEBUG_INVALIDPROPNAME
					, NULL);
					
			m_bRaiseException	= TRUE;	 
			return NULL;			
		}

		// Make sure we are not going over limit
	    if ((m_col.m_cElements + 1) > m_pDataObject->m_dwMaxProperties){     
			AddInDebugOutput(&m_excepinfo
					, TEXT("CProperties::Add")
					, IDS_DEBUG_COLLECTIONISFULL
					, NULL); 
			MYRAISEEXCEPTION;
    	}                       
		
		// Create and init new property
    	if (FAILED(CProperty::Create(Name.bstrVal, pXRTDispatch, &pProperty))){
			AddInDebugOutput(&m_excepinfo
					, TEXT("CProperties::Add")
					, IDS_DEBUG_CALLTOXRTPROPSADDFAILED
					, NULL); 
			MYRAISEEXCEPTION;
    	}       
        
        // Save properties in property
   		pProperty->m_pProperties = this;     
   	                                                    
        // Add property to cache
		if(m_pRequest && (int) m_col.m_cElements > m_pDataObject->m_bDontIncludeNames){	
			if(FAILED(m_pDataObject->m_pCache->AddProperty(m_pRequest
					, &pProperty->m_pxlProperty))){
				AddInDebugOutput(&m_excepinfo
					, TEXT("CProperties::Add")
					, IDS_DEBUG_CACHEADDPROPFAILED
					, NULL); 
				MYRAISEEXCEPTION;
			}
		}						   
        
        // Adds property to WOSA/XRT object and save LPDISPATCH
    	if (FAILED(XRTAdd(&Name, &pProperty->m_pXRTDispatch))){
			AddInDebugOutput(&m_excepinfo
				, TEXT("CProperties::Add")
				, IDS_DEBUG_CACHEADDPROPFAILED
				, NULL); 
			MYRAISEEXCEPTION;
    	}       

		// Internal QI; just a recast    
    	pProperty->QueryInterface(IID_IDispatch, (LPVOID FAR *) &pMyDispatch);

		// Saves ordinal index into collection
		pProperty->m_dwIndex = m_col.m_cElements;
		
        // AddObject to collection
    	if (FAILED(m_col.AddObject(pMyDispatch, Name.bstrVal, FALSE, 0))){
			AddInDebugOutput(&m_excepinfo
					, TEXT("CProperties::Add")
					, IDS_DEBUG_ADDOBJECTTOCOLLECTIONFAILED
					, NULL); 
			MYRAISEEXCEPTION;
    	}
        
        // Release from create
    	pProperty->Release();
    	pProperty = NULL;
        
        // Update property max count
		if(m_pDataObject && m_col.m_cElements > m_pDataObject->m_pRequests->m_dwMaxColCount){                                  
    		m_pDataObject->m_pRequests->m_dwMaxColCount = m_col.m_cElements;
	    }

	    return pMyDispatch;
	}
	    
	EXCEPT{
    	m_bRaiseException = TRUE;    

		//Clean Up
		if(pProperty)
			delete pProperty;

   		AddInDebugOutput(&m_excepinfo, TEXT("CProperties::Add")
   			, IDS_DEBUG_EXCEPTION, NULL); 

		m_bRaiseException = TRUE;

    	return NULL;   
	}
}

/*
 * CProperties::Add
 *
 * Purpose:
 *  Automation Add - Creates and adds Cproperties to collection
 *
 * Parameters:
 *  Name			Name of property to add
 *	ppProperty      Pointer to return created property
 *	ppNode     	 	Pointer to return created property's node
 *
 * Return Value:
 *  HRESULT
 *
 */

HRESULT CProperties::Add(VARIANT Name, CProperty FAR* FAR* ppProperty
		, CNode FAR* FAR* ppNode)
{
    LPDISPATCH pMyDispatch = NULL;
    CProperty FAR* pProperty = NULL;
    CNode FAR *pNode = NULL;

	TRY{          
		if(FAILED(CheckVariant(&Name, TRUE))){
    		AddInDebugOutput(NULL, TEXT("CProperties::Add")
    			, IDS_DEBUG_BADPOINTER, NULL); 
			return ResultFromScode(E_FAIL);
		}

		if(ppProperty != NULL){
			if(FAILED(CheckPointer((LPVOID) ppProperty, 0))){
    			AddInDebugOutput(NULL, TEXT("CProperties::Add")
    				, IDS_DEBUG_BADPOINTER, NULL); 
				return ResultFromScode(E_FAIL);
			}

		    *ppProperty = NULL;
		}

		if(pNode != NULL){
			if(FAILED(CheckPointer((LPVOID) ppNode, 0))){
    			AddInDebugOutput(NULL, TEXT("CProperties::Add")
    				, IDS_DEBUG_BADPOINTER, NULL); 
				return ResultFromScode(E_FAIL);
			}
	
	    	*ppNode = NULL;
		}
		
		// ake surre we are not going over limit
    	if (m_col.m_cElements + 1 > m_pDataObject->m_dwMaxProperties){     
			AddInDebugOutput(&m_excepinfo
					, TEXT("CProperties::Add")
					, IDS_DEBUG_COLLECTIONISFULL
					, NULL); 
			MYRAISEEXCEPTION;
    	}                       
		
		// Create new property
    	if (FAILED(CProperty::Create(Name.bstrVal, NULL, &pProperty))){
			AddInDebugOutput(&m_excepinfo
					, TEXT("CProperties::Add")
					, IDS_DEBUG_CALLTOXRTPROPSADDFAILED
					, NULL); 
			MYRAISEEXCEPTION;
    	}       
		
		// Save pointer to property's collection 	
   		pProperty->m_pProperties = this;     
         
		// Internal QI; just a recast    
    	pProperty->QueryInterface(IID_IDispatch, (LPVOID FAR *) &pMyDispatch);
        
        // Save ordinal index into collection
		pProperty->m_dwIndex = m_col.m_cElements;
    
    	// Add object to the collection
    	if(FAILED(m_col.AddObject(pMyDispatch, Name.bstrVal, FALSE, 0, &pNode))){
			AddInDebugOutput(&m_excepinfo
					, TEXT("CProperties::Add")
					, IDS_DEBUG_ADDOBJECTTOCOLLECTIONFAILED
					, NULL); 
			MYRAISEEXCEPTION;
    	}
        
        // Add property to cache
    	if(m_pRequest && (int) m_col.m_cElements > m_pDataObject->m_bDontIncludeNames){
			if(FAILED(m_pDataObject->m_pCache->AddProperty(m_pRequest
					, &pProperty->m_pxlProperty))){
				AddInDebugOutput(&m_excepinfo
					, TEXT("CProperties::Add")
					, IDS_DEBUG_CACHEADDPROPFAILED
					, NULL); 
				MYRAISEEXCEPTION;
			}
		}
		
		// Saved parent request
		pProperty->m_pRequest = m_pRequest;

		// I addref in constructer    
    	pMyDispatch->Release();
        pMyDispatch = NULL;
        
		//Store in collection until assigned to a property
		if(ppProperty)
	    	*ppProperty = pProperty;
        
		// If asked for, return its node
		if(ppNode)
       		*ppNode = pNode;

		return S_OK;		
	}

	EXCEPT{
		//Clean Up
		if(pProperty)
			delete pProperty;
	
    	m_bRaiseException = TRUE;    

   		AddInDebugOutput(&m_excepinfo, TEXT("CProperties::Add")
   			, IDS_DEBUG_EXCEPTION, NULL);
   			
		m_bRaiseException = TRUE;

   		return ResultFromScode(E_FAIL);
	}
}


/*
 * CProperties::Count
 *
 * Purpose:
 *  Return the number of properties in collection
 *
 * Return Value:
 *  long
 *
 */

STDMETHODIMP_(long)     
CProperties::get_Count(void)  
{
    return m_col.m_cElements;             
}


/*
 * CProperties::Item
 *
 * Purpose:
 *  Returns specific property from collection
 *
 * Return Value:
 *  LPDISPATCH
 *
 */

STDMETHODIMP_(LPDISPATCH)
CProperties::get_Item(VARIANT Index)
{
	LPDISPATCH pDispatch = NULL;
    HRESULT hr;
    
	TRY{
		if(FAILED(CheckVariant(&Index, TRUE))){
			AddInDebugOutput(&m_excepinfo
					, TEXT("CProperties::g_Item")
					, IDS_DEBUG_INVALIDINDEX
					, NULL); 
			m_bRaiseException = TRUE;
			return NULL;
		}
	
		// check index    
   		if(Index.vt == VT_I2){
   			hr = m_col.FindObjectByIndex((long) Index.iVal, &pDispatch);
   		}else if(Index.vt == VT_I4){
	   		hr = m_col.FindObjectByIndex(Index.lVal, &pDispatch);
   		}else if(Index.vt == VT_BSTR){
	   		hr = m_col.FindObject(Index.bstrVal, &pDispatch, NULL);
   		}else{
			AddInDebugOutput(&m_excepinfo
					, TEXT("CProperties::g_Item")
					, IDS_DEBUG_INVALIDINDEX
					, NULL); 
  			MYRAISEEXCEPTION;
	   	}

		// Find takes care of AddRef

    	return pDispatch;
	}
	
	EXCEPT{
		if(pDispatch)
			pDispatch->Release();
	
		m_bRaiseException = TRUE;

   		AddInDebugOutput(&m_excepinfo, TEXT("CProperties::AddExistingProps")
   			, IDS_DEBUG_EXCEPTION, NULL); 

		return NULL;
	}
}


/*
 * CProperties::get_NewEnum
 *
 * Purpose:
 *  Returns an enumerator (IEnumVARIANT) for the items curently in the collection. 
 *  The NewEnum property is restricted and so is invisible to users of an 
 *  automation controller's scripting language. Automation controllers that support
 *  a 'For Each' statement to iterate through the elements of a collection will use
 *  the enumerator returned by NewEnum. The enumerator creates a snapshot of the
 *  the current state of the collection.
 *
 */

STDMETHODIMP_(LPUNKNOWN)
CProperties::get__NewEnum(void)
{
    LPUNKNOWN punkEnumVariant = NULL;

	TRY{
		if(FAILED(m_col.GetEnum(&punkEnumVariant))){
			MYRAISEEXCEPTION;	
		}
		
    	return punkEnumVariant; 
	}
    
	EXCEPT{
    	m_bRaiseException = TRUE;    
		
   		AddInDebugOutput(&m_excepinfo, TEXT("CDataObjects::_NewEnum")
   			, IDS_DEBUG_EXCEPTION, NULL); 

		return NULL;    	
	}
}

/*
 * CProperties::Remove
 *
 * Purpose:
 *  Removes specified item from collection. Signals CProperties::Invoke to raise
 *  exception on ResultFromScode(E_FAIL).
 *
 * Parameters:   
 *   vIndex   Index of item to be removed. This index can be an integer or a string.
 *
 */

STDMETHODIMP_(void)
CProperties::Remove(VARIANT vParam)
{	
	ULONG lIndex;
	LPDISPATCH pDispatch = NULL;
	CProperty * pProperty = NULL;

	TRY{
		if(FAILED(CheckVariant(&vParam, TRUE))){
			AddInDebugOutput(&m_excepinfo
					, TEXT("CProperties::get_Item")
					, IDS_DEBUG_INVALIDINDEX
					, NULL); 
			m_bRaiseException = TRUE;
			return ;
		}
		
		// Remove proeprty from XRTObject
		if(FAILED(XRTRemove(&vParam))){
			AddInDebugOutput(&m_excepinfo
					, TEXT("CProperties::Remove")
					, IDS_DEBUG_CALLTOXRTPROPSREMOVE
					, NULL); 
  			MYRAISEEXCEPTION;
		}
		
		// Find object based on VARIANT index
   		if(vParam.vt == VT_I2){
   			hr = m_col.FindObjectByIndex((long) vParam.iVal, &pDispatch);
			lIndex = (long) vParam.iVal;
   		}else if(vParam.vt == VT_I4){
   			hr = m_col.FindObjectByIndex(vParam.lVal, &pDispatch);
			lIndex = (long) vParam.lVal;
   		}else if(vParam.vt == VT_BSTR){
   			hr = m_col.FindObject(vParam.bstrVal, &pDispatch, &lIndex);
   		}else{
			AddInDebugOutput(&m_excepinfo
					, TEXT("CProperties::Remove")
					, IDS_DEBUG_INVALIDINDEX
					, NULL); 
  	
  			MYRAISEEXCEPTION;
   		}

		if(FAILED(hr)){        
			AddInDebugOutput(&m_excepinfo
					, TEXT("CProperties::Remove")
					, IDS_DEBUG_FINDOBJECTINCOLLECTIONFAILED
					, NULL); 
  			MYRAISEEXCEPTION;
		}
		
		// Recast to CProperty
		// Internal QI just recasts		
		pDispatch->QueryInterface(IID_IProperty, (LPVOID *) &pProperty);
		
		pProperty->m_pxlProperty->xltype = xltypeNil;
        
		// Update collection
		if(FAILED(m_col.RemoveObject(pDispatch))){
			AddInDebugOutput(&m_excepinfo
					, TEXT("CPropertiess::Remove")
					, IDS_DEBUG_REMOVEOBJECTFROMCOLLECTIONFAILED
					, NULL); 
  			MYRAISEEXCEPTION;
		}
		
		//Release from FindObject
		pDispatch->Release();
		pDispatch = NULL;
        
        // Release from QI
		pProperty->Release();
		pProperty = NULL;
	
	    return; 
	}

	EXCEPT{
		if(pDispatch)
			pDispatch->Release();
						
		if(pProperty)
			pProperty->Release();
						
		m_bRaiseException = TRUE;

   		AddInDebugOutput(&m_excepinfo, TEXT("CProperties::Remove")
   			, IDS_DEBUG_EXCEPTION, NULL); 

		return;
	}
}

/*
 * CProperties::Object - Read-only property
 *
 * Purpose:
 *	Returns WOSA/XRT object CProperty object. Used by VBA developer to call
 *	properties and methods which are specific to the WOSA/XRT object.
 *
 * Return Value:
 *  VARIANT
 *
 */

STDMETHODIMP_(IDispatch *)
CProperties::get_Object()
{
	if(FAILED(CheckDispatch(m_pXRTDispatch))){
   		AddInDebugOutput(NULL, TEXT("CDataObject::Create")
   			, IDS_DEBUG_BADPOINTERTOXRTOBJECT, NULL); 

		m_bRaiseException = TRUE;
		return NULL;
	}

	m_pXRTDispatch->AddRef();	
                   
	return m_pXRTDispatch;                     
}

/*
 * CProperties::DataObject
 *
 * Purpose:
 *	returns pointer back to parent DataObject
 *
 * Return Value:
 *  LPDISPATCH
 *
 */

STDMETHODIMP_(IDispatch *)
CProperties::get_DataObject(){                                                                            
	LPDISPATCH pDispatch = NULL;
	
	m_pDataObject->QueryInterface(IID_IDispatch, (LPVOID FAR *) &pDispatch);
	
	return pDispatch;
}


/*
 * CProperties::Reset
 *
 * Purpose:
 *	Resets collection to zero
 *
 * Return Value:
 *  void
 *
 */

STDMETHODIMP_(void) CProperties::Reset(void){
	static DISPID dispid = -1;
	static DISPPARAMS dispparams = {NULL, NULL, 0, 0};
	static TCHAR szMember[] = TEXT("Reset");
	VARIANT vResult, v;
	CProperty FAR *pProperty  = NULL;

	TRY{
		VariantInit(&vResult);
		VariantInit(&v);
	
		if(FAILED(InvokeMethodHelper(m_pXRTDispatch, &dispid, 
				&dispparams, szMember, NULL))){

			AddInDebugOutput(&m_excepinfo
					, TEXT("CProperties::Reset")
					, IDS_DEBUG_CALLTOXRTPROPSRESETFAILED
					, NULL); 
			MYRAISEEXCEPTION;
		}
                
		if(FAILED(m_col.RemoveAll())){
			AddInDebugOutput(&m_excepinfo
					, TEXT("CProperties::Reset")
					, IDS_DEBUG_REMOVEALLFROMCOLLECTIONFAILED
					, NULL); 
			MYRAISEEXCEPTION;
		}

		return;
	}

	EXCEPT{
   		AddInDebugOutput(&m_excepinfo, TEXT("CProperties::Reset")
   			, IDS_DEBUG_EXCEPTION, NULL); 

		m_bRaiseException = TRUE;

		return;		
	}
}

/*
 * CProperties::XRTAdd
 *
 * Purpose:
 *	Calls Properties.Add for real WOSA/XRT object
 *
 * Return Value:
 *  void
 *
 */

HRESULT CProperties::XRTAdd(VARIANT FAR *pvName, LPDISPATCH FAR *ppDispatch){
	static DISPID dispid = -1;
	static TCHAR szMember[] = TEXT("Add");
	VARIANT param;
	DISPPARAMS dispparams;
	VARIANT vResult;

	TRY{
		if(FAILED(CheckVariant(pvName, TRUE))){
			return ResultFromScode(E_FAIL);
		}

		if(FAILED(CheckPointer((LPVOID) ppDispatch, 0))){
    		AddInDebugOutput(NULL, TEXT("CRequests::XRTAdd")
    			, IDS_DEBUG_BADPOINTER, NULL); 
			return ResultFromScode(E_FAIL);
		}
		
		*ppDispatch = NULL;
	
		VariantInit(&vResult);			
		VariantInit(&param);			

		if(FAILED(VariantCopy((VARIANT *) &param, pvName))){
   			AddInDebugOutput(NULL, TEXT("CRequests::XRTAdd")
    			, IDS_DEBUG_VARIANTCOPYFAILED, NULL); 
			return ResultFromScode(E_FAIL);
		}
		
		dispparams.rgvarg = (VARIANT *) &param;
		dispparams.rgdispidNamedArgs = NULL;
		dispparams.cArgs = 1;
		dispparams.cNamedArgs = 0;

		if(FAILED(InvokeMethodHelper(m_pXRTDispatch, &dispid, 
				&dispparams, szMember, &vResult))){

	   		AddInDebugOutput(&m_excepinfo, TEXT("CRequests::XRTAdd")
   				, IDS_DEBUG_EXCEPTION, NULL); 

			MYRAISEEXCEPTION;
		}

		if(FAILED(CheckVariant(&vResult, TRUE))){
	   		AddInDebugOutput(&m_excepinfo, TEXT("CRequests::XRTAdd")
   				, IDS_DEBUG_INVALIDRETURNVALUEONXRTMETHOD, NULL); 
			MYRAISEEXCEPTION;
		}

		*ppDispatch = vResult.pdispVal;

		// I am not sure if you need this addref.
		// Your WOSA/XRT object should do it for you
		(*ppDispatch)->AddRef();

		VariantClear((VARIANT *)&param);			

		return S_OK;
	}

	EXCEPT{
		VariantClear((VARIANT *)&param);			
		VariantClear((VARIANT *)&vResult);			

   		AddInDebugOutput(&m_excepinfo, TEXT("CRequests::XRTAdd")
   			, IDS_DEBUG_EXCEPTION, NULL); 

		return ResultFromScode(E_FAIL);
	}
}

/*
 * CProperties::XRTRemove
 *
 * Purpose:
 *	calls Properties.remove on real WOSA/XRT object
 *
 * Return Value:
 *  void
 *
 */

HRESULT CProperties::XRTRemove(VARIANT FAR *pvIndex){
	static DISPID dispid = -1;
	static TCHAR szMember[] = TEXT("Remove");
	VARIANT param;
	DISPPARAMS dispparams;
	VARIANT vResult;

	TRY{    
		if(FAILED(CheckVariant(pvIndex, TRUE))){
			return ResultFromScode(E_FAIL);
		}
	
		VariantInit(&param);
		VariantInit(&vResult);

		if(FAILED(VariantCopy(&param, pvIndex))){
   			AddInDebugOutput(NULL, TEXT("CRequests::XRTAdd")
    			, IDS_DEBUG_INVALIDPARAMETERS, NULL); 

			return ResultFromScode(E_FAIL);
		}
		
		dispparams.rgvarg = &param;
		dispparams.rgdispidNamedArgs = NULL;
		dispparams.cArgs = 1;
		dispparams.cNamedArgs = 0;
	
		if(FAILED(InvokeMethodHelper(m_pXRTDispatch, &dispid, 
				&dispparams, szMember, &vResult))){

			VariantClear(&param);

	   		AddInDebugOutput(&m_excepinfo, TEXT("CRequests::XRTRemove")
   				, IDS_DEBUG_EXCEPTION, NULL); 

			return ResultFromScode(E_FAIL);
		}
		
		VariantClear(&param);
	
		return S_OK;
	}

	#ifdef _WIN32
	EXCEPT{
		VariantClear(&param);
		VariantClear(&vResult);

   		AddInDebugOutput(&m_excepinfo, TEXT("CRequests::XRTRemove")
   			, IDS_DEBUG_EXCEPTION, NULL); 

		return ResultFromScode(E_FAIL);;
	}
	#endif
}


/*
 * CProperties::XRTItem 
 *
 * Purpose:
 *	XRTItem calls Properties.Item on real WOSA/XRT object
 *
 * Return Value:
 *  VARIANT
 *
 */

HRESULT CProperties::XRTItem(VARIANT FAR *pvIndex, LPDISPATCH FAR *ppDispatch){
	static DISPID dispid = -1;
	VARIANTARG params[1];
	DISPPARAMS dispparams;
	TCHAR szMember[] = TEXT("Item");
	VARIANT vResult;

	TRY{
		if(FAILED(CheckVariant(pvIndex, TRUE))){
			return ResultFromScode(E_FAIL);
		}

		if(FAILED(CheckPointer((LPVOID) ppDispatch, 0))){
    		AddInDebugOutput(NULL, TEXT("CRequests::XRTItem")
    			, IDS_DEBUG_BADPOINTER, NULL); 
			return ResultFromScode(E_FAIL);
		}

		*ppDispatch = NULL;
			
		VariantInit(&vResult);			
		VariantInit(&params[0]);			
		
		if(FAILED(VariantCopy(&params[0], pvIndex))){
   			AddInDebugOutput(NULL, TEXT("CRequests::XRTItem")
    			, IDS_DEBUG_VARIANTCOPYFAILED, NULL); 
			return ResultFromScode(E_FAIL);
		}

		dispparams.rgvarg = params;
		dispparams.rgdispidNamedArgs = NULL;
		dispparams.cArgs = 1;
		dispparams.cNamedArgs = 0;

		if(FAILED(InvokePropGetHelper(m_pXRTDispatch, &dispid, 
				&dispparams, szMember, &vResult))){
	   		AddInDebugOutput(&m_excepinfo, TEXT("CRequests::XRTItem")
   				, IDS_DEBUG_INVALIDRETURNVALUEONXRTMETHOD, NULL); 
			MYRAISEEXCEPTION;
		}		

		if(FAILED(CheckVariant(&vResult, TRUE))){
	   		AddInDebugOutput(&m_excepinfo, TEXT("CRequests::XRTItem")
   				, IDS_DEBUG_INVALIDRETURNVALUEONXRTMETHOD, NULL); 
			MYRAISEEXCEPTION;
		}

		*ppDispatch = vResult.pdispVal;
		// I am not sure if you need this addref.
		// Your WOSA/XRT object should do it for you
		(*ppDispatch)->AddRef();


		return S_OK;
	}
	
	EXCEPT{
		VariantClear(&vResult);			
		VariantClear(&params[0]);			

		return ResultFromScode(E_FAIL);	
	}
}
