// 	WOSA/XRT Excel 5.0 AddIn Version 1.00
//
//	COPYRIGHT:
//
//  	(C) Copyright Microsoft Corp. 1994.  All rights reserved.
//
//   	You have a royalty-free right to use, modify, reproduce and
//   	distribute the Sample Files (and/or any modified version) in
//   	any way you find useful, provided that you agree that
//   	Microsoft has no warranty obligations or liability for any
//   	Sample Application Files which are modified.
//
//
// 	Revision History
//    	October 15, 1994   rc    Version 1.00 Release
//
// 	Comments:
// 		The CRequest class implements the IUnknown and IDispatch interfaces.
// 		The IDispatch interface supersets all the WOSA/XRT Request methods 
// 		and properties. The other properties and methods are unique to to add-in
// 		and are defined throughout this module... 

#include "precomp.h"
#include "guid.h"

#include "dataobj.h"
#include "requests.h"
#include "request.h"
#include "props.h"
          
/*
 * CRequest::Create
 *
 * Purpose:
 *  Creates an instance of the client object and initializes it.
 *
 * Parameters:
 *  ppRequest    Returns Item object.
 *
 * Return Value:
 *  HRESULT
 *
 */ 
 

HRESULT CRequest::Create(BSTR bstrName, CRequest FAR* FAR* ppRequest) 
{   
    CRequest FAR* pRequest = NULL;
	
	TRY{
		if(FAILED(CheckPointer((LPVOID) ppRequest, 0))){
    		AddInDebugOutput(NULL, TEXT("CRequest::Create")
    			, IDS_DEBUG_BADPOINTER, NULL); 
			return ResultFromScode(E_FAIL);
		}
	
	    *ppRequest = NULL;

    	pRequest = new CRequest();
    	
	    if (pRequest == NULL){
			AddInDebugOutput(&pRequest->m_excepinfo
					, TEXT("CRequest::Create")
					, IDS_DEBUG_REQUESTALLOCFAILED
					, NULL); 
			return ResultFromScode(E_FAIL);
    	}                                                                         
    
    	// Load type information for the collection from type library. CreateStdDispatch
    	// is not used to create a standard IDispatch implemenation because this
    	// collection implementation raises exceptions on ResultFromScode(E_FAIL)s.    
	    if (FAILED(LoadTypeInfo(&pRequest->m_ptinfo, IID_IRequest))){
			AddInDebugOutput(&pRequest->m_excepinfo
					, TEXT("CRequest::Create")
					, IDS_DEBUG_COULDNOTLOADTYPEINFO
					, NULL); 
			MYRAISEEXCEPTION;
    	}
                   
		pRequest->m_bstrName = SysAllocString(bstrName);
    
    	*ppRequest = pRequest;       
    
    	return S_OK;
	}

	EXCEPT{
   		AddInDebugOutput(NULL, TEXT("CRequest::Create")
   			, IDS_DEBUG_EXCEPTION, NULL); 

    	if(pRequest) 
    		delete pRequest;

	    return ResultFromScode(E_FAIL);
	}
}

/*
 * CRequest::CRequest
 *
 * Purpose:
 *  Constructor for CRequest object. Initializes members to NULL.
 *
 */                             
 
CRequest::CRequest()
{
	m_dispProperties = NULL;
	m_pProperties = NULL;
	m_pXRTDispatch = NULL;
	m_pxlRequest = NULL;
	m_pDataObject = NULL;
	m_dwIndex = 0;
	m_bstrName = NULL;
	m_ptinfo = NULL;
	m_cRef = 1;
   	m_bRaiseException = 0;
	m_bstrPassThroughFunc = NULL;
}

/*
 * CRequest::~CRequest
 *
 * Purpose:
 *  Destructor for CRequest object. 
 *
 */

CRequest::~CRequest()
{    
	if(SUCCEEDED(CheckPointer(m_pProperties)))
		m_pProperties->Release();
	
	if(SUCCEEDED(CheckPointer(m_bstrName)))
		SysFreeString(m_bstrName);                        

	if(SUCCEEDED(CheckDispatch(m_pXRTDispatch)))
		m_pXRTDispatch->Release();
    
	if(SUCCEEDED(CheckPointer(m_ptinfo)))
   		m_ptinfo->Release();

	if(SUCCEEDED(CheckPointer(m_bstrPassThroughFunc)))
		SysFreeString(m_bstrPassThroughFunc);
}                                                            																																															


/*
 * CRequest::QueryInterface, AddRef, Release
 *
 * Purpose:
 *  Implements IUnknown::QueryInterface, AddRef, Release
 *
 */
 
STDMETHODIMP
CRequest::QueryInterface(REFIID iid, void FAR* FAR* ppv) 
{   
    *ppv = NULL;
	
    if (iid == IID_IUnknown) 
		*ppv = this;  
    else if (iid == IID_IRequest) 
		*ppv = this;    
    else if (iid == IID_IDispatch || iid == IID_DRequest)
		*ppv = this;   
    else 
    	return ResultFromScode(E_NOINTERFACE); 

    AddRef();
    return S_OK;    
}


STDMETHODIMP_(ULONG)
CRequest::AddRef(void)
{ 
    return ++m_cRef;
}


STDMETHODIMP_(ULONG)
CRequest::Release(void)
{ 
    if(--m_cRef == 0){
		delete this;
		return 0;
    }
    return m_cRef;
} 


/*
 * CRequest::GetTypeInfoCount
 *
 * Purpose:
 *  Implements IDispatch::GetTypeInfoCount.
 *
 */
STDMETHODIMP
CRequest::GetTypeInfoCount(UINT FAR* pctinfo)
{
    *pctinfo = 1;
    return S_OK;
}

/*
 * CRequest::GetTypeInfo
 *
 * Purpose:
 *  Implements IDispatch::GetTypeInfo. 
 *
 */
STDMETHODIMP
CRequest::GetTypeInfo(
      UINT itinfo,
      LCID lcid,
      ITypeInfo FAR* FAR* pptinfo)
{    
    *pptinfo = NULL;
     
    if(itinfo != 0)
		return ResultFromScode(DISP_E_BADINDEX);
    
    m_ptinfo->AddRef(); 
    *pptinfo = m_ptinfo;
    
    return S_OK;
}

/*
 * CRequest::GetIDsOfNames
 *
 * Purpose:
 *  Implements IDispatch::GetIDsOfNames.  The standard implementation, DispGetIDsOfNames,
 *  is used.
 *
 */
STDMETHODIMP 
CRequest::GetIDsOfNames(
      REFIID riid,
      OLECHAR FAR* FAR* rgszNames,
      UINT cNames,
      LCID lcid,
      DISPID FAR* rgdispid)
{
    return DispGetIDsOfNames(m_ptinfo, rgszNames, cNames, rgdispid);
}

/*
 * CRequest::Invoke
 *
 * Purpose:
 *  Implements IDispatch::Invoke.  The standard implementation, DispInvoke,
 *  is used. Properties and methods exposed by the collection object will
 *  set m_bRaiseException to raise an exception. 
 *
 *	In Excel RealTime Extensions version 1.00, I could not get automation
 *	exception exceptions to occurr consistently with descriptions and 
 *  wCodes. 
 *
 */

STDMETHODIMP
CRequest::Invoke(
      DISPID dispidMember,
      REFIID riid,
      LCID lcid,
      WORD wFlags,
      DISPPARAMS FAR* pdispparams,
      VARIANT FAR* pvarResult,
      EXCEPINFO FAR* pexcepinfo,
      UINT FAR* puArgErr)
{  
    HRESULT hr = S_OK;
     
    if (NULL == pexcepinfo)
		return ResultFromScode(E_INVALIDARG); 

   	m_bRaiseException = FALSE;
	m_bstrPassThroughFunc = NULL;	 

   	if(FAILED(DispInvoke(this, m_ptinfo,dispidMember, wFlags, pdispparams,
			pvarResult, pexcepinfo, puArgErr))){
	
		pexcepinfo->bstrSource = SysAllocString(g_szTitle); 
		pexcepinfo->bstrDescription = SysAllocString(TEXT("Internal invoke error. Type library may be corrupt.")); 
       	pexcepinfo->wCode = 1170;

		PrepareExcepInfo(pexcepinfo);

		return ResultFromScode(DISP_E_EXCEPTION); 
	}
	    
	if(m_bstrPassThroughFunc != NULL){
		DISPID xrtdispid = -1;

		if(FAILED(CheckDispatch(m_pXRTDispatch))){
    		AddInDebugOutput(NULL, TEXT("CDataObject::Create")
    			, IDS_DEBUG_BADPOINTERTOXRTOBJECT, NULL); 
			return ResultFromScode(E_FAIL);
		}

		if(FAILED(m_pXRTDispatch->GetIDsOfNames(IID_NULL
			, (LPTSTR FAR *) &m_bstrPassThroughFunc, 1, lcid, &xrtdispid))){

			pexcepinfo->bstrSource = SysAllocString(g_szTitle); 
			pexcepinfo->bstrDescription = SysAllocString(TEXT("Dispatch call to bound XRT object failed.")); 
	       	pexcepinfo->wCode = 1171;

			PrepareExcepInfo(pexcepinfo);

			return ResultFromScode(DISP_E_EXCEPTION); 
		}

		SysFreeString(m_bstrPassThroughFunc);
		m_bstrPassThroughFunc = NULL;

		if(FAILED(m_pXRTDispatch->Invoke(xrtdispid, riid
			, lcid, wFlags, pdispparams, pvarResult
			, pexcepinfo , puArgErr))){

			pexcepinfo->bstrSource = SysAllocString(g_szTitle); 
			pexcepinfo->bstrDescription = SysAllocString(TEXT("Dispatch call to bound XRT object failed.")); 
	       	pexcepinfo->wCode = 1171;

			PrepareExcepInfo(pexcepinfo);
	
			return ResultFromScode(DISP_E_EXCEPTION); 
		}

		return S_OK;
	}

   	// Check if property accessor function or method called by DispInvoke wants to raise
   	// an exception. If so, return DISP_E_EXCEPTION after filling out the EXCEPINFO 
   	// structure. The scode and bstrDescription fields of EXCEPINFO are filled by the 
   	// property or method.    
   	if (m_bRaiseException){
		pexcepinfo->bstrDescription = m_excepinfo.bstrDescription;
		pexcepinfo->bstrSource = SysAllocString(g_szTitle);   
		pexcepinfo->wCode = m_excepinfo.wCode;   

		PrepareExcepInfo(pexcepinfo);

		m_bRaiseException = FALSE;
       	m_excepinfo.scode = NULL;
       	m_excepinfo.wCode = 0;

		return ResultFromScode(DISP_E_EXCEPTION); 
   	}else{
   		return hr;    
	}
}


/*
 * CRequest::Object - Read-only property
 *
 * Purpose:
 *	Returns WOSA/XRT object CProperty object. Used by VBA developer to call
 *	properties and methods which are specific to the WOSA/XRT object.
 *
 * Return Value:
 *  VARIANT
 *
 */

STDMETHODIMP_(IDispatch *)
CRequest::get_Object(){
	if(FAILED(CheckDispatch(m_pXRTDispatch))){
   		AddInDebugOutput(NULL, TEXT("CDataObject::Create")
   			, IDS_DEBUG_BADPOINTERTOXRTOBJECT, NULL); 

		m_bRaiseException = TRUE;
		return NULL;
	}

	m_pXRTDispatch->AddRef();	
                   
	return m_pXRTDispatch;                     
}          

/*
 * CRequest::Properties - Read-only property
 *
 * Purpose:
 * 	Return the properties collection of this request
 *
 * Return Value:
 *  VARIANT
 *
 */

STDMETHODIMP_(IDispatch *) 
CRequest::get_Properties(void){
	LPDISPATCH pDispatch = NULL;
	
	m_pProperties->QueryInterface(IID_IDispatch, (LPVOID FAR *) &pDispatch);
	
	return pDispatch;
}

/*
 * CRequest::Request - Read/Write property
 *
 * Purpose:
 * 	Specifies request of request
 *
 * Return Value:
 *  VARIANT
 *
 */

STDMETHODIMP_(BSTR)
CRequest::get_Request(void){
	// This property or method is an example of an XRT member which the
	// add-in does not need to be aware of. We can simply passes the 
	// call onto the XRT object. I do this in a centralized location:
	// ::Invoke.

	m_bstrPassThroughFunc = SysAllocString(TEXT("Request"));

	return NULL;		
}

STDMETHODIMP_(void)
CRequest::put_Request(BSTR newval){
	// This property or method is an example of an XRT member which the
	// add-in does not need to be aware of. We can simply passes the 
	// call onto the XRT object. I do this in a centralized location:
	// ::Invoke.

	m_bstrPassThroughFunc = SysAllocString(TEXT("Request"));

	return;		
}


/*
 * CRequest::XRTProperties - Read-only property
 *
 * Purpose:
 * 	Calls real WOSA/XRT object's Properties property
 *
 * Return Value:
 *  VARIANT
 *
 */

HRESULT CRequest::XRTProperties(LPDISPATCH FAR *ppDispatch){
	static DISPID dispid = -1;
	static DISPPARAMS dispparams = {NULL, NULL, 0, 0};
	static TCHAR szMember[] = TEXT("Properties");
	VARIANT vResult;

	TRY{	
		if(FAILED(CheckPointer((LPVOID) ppDispatch, 0))){
    		AddInDebugOutput(NULL, TEXT("CRequest::XRTProperties")
    			, IDS_DEBUG_BADPOINTER, NULL); 
			return ResultFromScode(E_FAIL);
		}
	
    	*ppDispatch = NULL;
    
		VariantInit(&vResult);
	
		if(FAILED(InvokePropGetHelper(m_pXRTDispatch, &dispid, 
				&dispparams, szMember, &vResult))){

			AddInDebugOutput(&m_excepinfo
				, TEXT("CRequest::XRTProperties")
				, IDS_DEBUG_CALLTOXRTREQUESTPROPERTIESFAILED
				, NULL); 

    	    MYRAISEEXCEPTION;
		}
		
		if(FAILED(CheckVariant(&vResult, FALSE))){
			AddInDebugOutput(&m_excepinfo
				, TEXT("CRequest::XRTProperties")
				, IDS_DEBUG_BADVARIANT
				, NULL); 

    	    MYRAISEEXCEPTION;
		}
        
        *ppDispatch = vResult.pdispVal;
		// I am not sure if you need this addref.
		// Your WOSA/XRT object should do it for you
		(*ppDispatch)->AddRef();

		return S_OK;
	}		
	
	EXCEPT{
		AddInDebugOutput(&m_excepinfo
			, TEXT("CRequest::XRTProperties")
			, IDS_DEBUG_EXCEPTION
			, NULL); 

		VariantClear(&vResult);
   	    
   	    return ResultFromScode(E_FAIL);
	}
	
}

