// 	WOSA/XRT Excel 5.0 AddIn Version 1.00
//
//	COPYRIGHT:
//
//  	(C) Copyright Microsoft Corp. 1994.  All rights reserved.
//
//   	You have a royalty-free right to use, modify, reproduce and
//   	distribute the Sample Files (and/or any modified version) in
//   	any way you find useful, provided that you agree that
//   	Microsoft has no warranty obligations or liability for any
//   	Sample Application Files which are modified.
//
//
// 	Revision History
//    	October 15, 1994   rc    Version 1.00 Release
//
// 	Comments:
//		Implementation of IClassFactory

#include "precomp.h"
#include "realtime.h" 
#include "xrtcf.h"

/*
 * CXrtCF::CXrtCF
 *
 * Purpose:
 *  Constructor
 *
 */

CXrtCF::CXrtCF(void)
{    
    m_cRef = 1; 
}

/*
 * CXrtCF::QueryInterface, AddRef, Release
 *
 * Purpose:
 *  Implements IUnknown::QueryInterface, AddRef, Release
 *
 */

STDMETHODIMP
CXrtCF::QueryInterface(REFIID iid, void FAR* FAR* ppv) 
{   
    *ppv = NULL;
    
    if (iid == IID_IUnknown || iid == IID_IClassFactory)
        *ppv = this;
    else 
        return ResultFromScode(E_NOINTERFACE); 

    AddRef();
    return S_OK;    
}


STDMETHODIMP_(ULONG)
CXrtCF::AddRef(void)
{
    return ++m_cRef;
}


STDMETHODIMP_(ULONG)
CXrtCF::Release(void)
{
    if(--m_cRef == 0)
    {
        delete this;
        return 0;
    }
    return m_cRef;
}

/*
 * CXrtCF::CreateInstance, LockServer
 *
 * Purpose:
 *  Implements IClassFactory::CreateInstance, LockServer
 *
 */ 
 
STDMETHODIMP
CXrtCF::CreateInstance(IUnknown FAR* punkOuter,
                         REFIID riid,
                         void FAR* FAR* ppv)
{
	TRY{                       
		g_pRealTimeExt = NULL;
	
	    *ppv = NULL;

    	// This implementation does not allow aggregation
    	if (punkOuter)
        	return ResultFromScode(CLASS_E_NOAGGREGATION);   

	   	if (FAILED(CRealTimeExt::Create(&g_pRealTimeExt))){
			AddInDebugOutput(NULL
					, TEXT("IClassFactory::CreateInstance")
					, IDS_DEBUG_CREATEREALTIMEEXTFAILED
					, NULL); 
  			MYRAISEEXCEPTION;
		}
        
	    if (FAILED(g_pRealTimeExt->QueryInterface(riid, ppv))){ 
			AddInDebugOutput(NULL
					, TEXT("IClassFactory::CreateInstance")
					, IDS_DEBUG_QIFORSPECIFIEDINTERFACEFAILED
					, NULL); 
  			MYRAISEEXCEPTION;
	    } 
    
		g_pRealTimeExt->Release();
        
    	return S_OK;
	}

	EXCEPT{
   		AddInDebugOutput(NULL, TEXT("IClassFactory::CreateInstance")
   			, IDS_DEBUG_EXCEPTION, NULL); 
   			
		if(g_pRealTimeExt)
        	delete g_pRealTimeExt;
        
    	return ResultFromScode(E_FAIL);
	}
}                

STDMETHODIMP
CXrtCF::LockServer(BOOL fLock)
{
    if(fLock){
		g_cObjCount++;    
    }else{
		g_cObjCount--;    
    }

    return S_OK;            
}


